#include "component_action_spawn.h"

/* 
	DEPENDENCIES:
	EntityCupper
	EntityChip
	EntityCustarpedo
	EntityCrumbler
	EntityChipper
	EntityExcorsant
*/

static GameEntityType _game_entity_type_get(ComponentActionSpawn* self, ECS* ecs);
static void _spawn_timer_set(ComponentActionSpawn* self, ECS* ecs);
static void _spawn_entity(ComponentActionSpawn* self, ECS* ecs);
static void _spawn_entity_out_of_bounds_position_get(ComponentActionSpawn* self, ECS* ecs, GameEntityType type, vec3 position, vec2 size, u32 id, SpawnDirection* direction);
static void _spawn_entity_in_bounds_position_get(ComponentActionSpawn* self, ECS* ecs, GameEntityType type, vec3 position, vec2 size, u32 id);
static void _alert_no_direction_spawn(ComponentActionSpawn* self, ECS* ecs, vec3 position, IconType type);
static void _alert_direction_spawn(ComponentActionSpawn* self, ECS* ecs, vec3 position, SpawnDirection direction, IconType type);

/* Spawns an alert notice. */
static void
_alert_no_direction_spawn(ComponentActionSpawn* self, ECS* ecs, vec3 position, IconType type)
{
	vec3 alertPosition;
	vec3 alertIconPosition;

	glm_vec3_copy(position, alertPosition);

	alertPosition[0] = CLAMP(alertPosition[0], COMPONENT_ACTION_SPAWN_ALERT_X_POSITION_MIN, COMPONENT_ACTION_SPAWN_ALERT_X_POSITION_MAX);
	alertPosition[1] = CLAMP(alertPosition[1], COMPONENT_ACTION_SPAWN_ALERT_Y_POSITION_MIN, COMPONENT_ACTION_SPAWN_ALERT_Y_POSITION_MAX);
	alertPosition[2] = COMPONENT_ACTION_SPAWN_ALERT_Z_OFFSET;

	glm_vec3_copy(alertPosition, alertIconPosition);

	alertIconPosition[1] += COMPONENT_ACTION_SPAWN_ALERT_OFFSET;

	entity_alert_init
	(
		ecs,
		ecs_entity_add(ecs),
		alertPosition
	);

	entity_alert_icon_init
	(
		ecs,
		ecs_entity_add(ecs),
		alertIconPosition,
		type
	);
}

/* Spawns an alert notice. */
static void
_alert_direction_spawn(ComponentActionSpawn* self, ECS* ecs, vec3 position, SpawnDirection direction, IconType type)
{
	vec3 alertPosition;
	vec3 alertIconPosition;

	glm_vec3_copy(position, alertPosition);

	switch (direction)
	{
		case SPAWN_DIRECTION_NORTH:
		case SPAWN_DIRECTION_SOUTH:
			alertPosition[0] = position[0];
			break;
		case SPAWN_DIRECTION_EAST:
		case SPAWN_DIRECTION_WEST:
			alertPosition[1] = position[1];
			break;
		default:
			break;
	}

	alertPosition[0] = CLAMP(alertPosition[0], COMPONENT_ACTION_SPAWN_ALERT_X_POSITION_MIN, COMPONENT_ACTION_SPAWN_ALERT_X_POSITION_MAX);
	alertPosition[1] = CLAMP(alertPosition[1], COMPONENT_ACTION_SPAWN_ALERT_Y_POSITION_MIN, COMPONENT_ACTION_SPAWN_ALERT_Y_POSITION_MAX);
	alertPosition[2] = COMPONENT_ACTION_SPAWN_ALERT_Z_OFFSET;
	
	glm_vec3_copy(alertPosition, alertIconPosition);

	switch (direction)
	{
		case SPAWN_DIRECTION_NORTH:
			alertIconPosition[1] += COMPONENT_ACTION_SPAWN_ALERT_OFFSET;
			break;
		case SPAWN_DIRECTION_SOUTH:
			alertIconPosition[1] -= COMPONENT_ACTION_SPAWN_ALERT_OFFSET;
			break;
		case SPAWN_DIRECTION_EAST:
			alertIconPosition[0] -= COMPONENT_ACTION_SPAWN_ALERT_OFFSET;
			break;
		case SPAWN_DIRECTION_WEST:
			alertIconPosition[0] += COMPONENT_ACTION_SPAWN_ALERT_OFFSET;
			break;
		default:
			break;
	}

	entity_alert_init
	(
		ecs,
		ecs_entity_add(ecs),
		alertPosition
	);

	entity_alert_icon_init
	(
		ecs,
		ecs_entity_add(ecs),
		alertIconPosition,
		type
	);
}

/* Sets the spawn timer. */
static void
_spawn_timer_set(ComponentActionSpawn* self, ECS* ecs)
{	
	ComponentLevel* level;
	u32 usedLevel;

	level = ecs_component_get(ecs, ECS_COMPONENT_LEVEL, self->component.id);

	usedLevel = level->value + self->levelBonus;

	self->timer = COMPONENT_ACTION_SPAWN_TIMER_BASE - (COMPONENT_ACTION_SPAWN_TIMER_LEVEL_MULTIPLIER * (usedLevel - 1));

	self->timer = MIN(self->timer, COMPONENT_ACTION_SPAWN_TIMER_MINIMUM);
}

/* Based on spawn rarity, randomly select an entity type to spawn. */
static GameEntityType
_game_entity_type_get(ComponentActionSpawn* self, ECS* ecs)
{
	GameEntityType type;
	Vector selectTypes;

	/* No point in doing this with only one type. */
	if (self->types.count == 1)
	{
		type = *(GameEntityType*)vector_get(&self->types, 0);
		return type;
	}

	vector_init(&selectTypes, sizeof(GameEntityType));

	/* Entities have different spawn rates. */
	/* Populate a list based on frequency (i.e., SPAWN_RARITY_COMMON types will appear 3x
	 * more likely than SPAWN_RARITY_RARE). */
	for (s32 i = 0; i < (s32)self->types.count; i++)
	{
		SpawnRarityType spawnRarityType;
		GameEntityType gameEntityType;
		u32 spawnCheckCount;

		gameEntityType = *(GameEntityType*)vector_get(&self->types, i);

		spawnRarityType = SPAWN_GAME_ENTITY_RARITY_TYPES[gameEntityType];

		spawnCheckCount = (SPAWN_RARITY_COUNT - spawnRarityType) + 1;

		for (s32 j = 0; j < (s32)spawnCheckCount; j++)
			vector_push(&selectTypes, &gameEntityType);
	}

	type = *(GameEntityType*)vector_get(&selectTypes, RANDOM_S32(0, selectTypes.count - 1));

	vector_free(&selectTypes);
	
	return type;
}

/* Spawns an entity out of bounds. */
static void
_spawn_entity_out_of_bounds_position_get(ComponentActionSpawn* self, ECS* ecs, GameEntityType type, vec3 position, vec2 size, u32 id, SpawnDirection* direction)
{
	*direction = RANDOM_S32(SPAWN_DIRECTION_NORTH, SPAWN_DIRECTION_WEST);

	switch (*direction)
	{
		case SPAWN_DIRECTION_NORTH:
			position[0] = RANDOM_F32(SPAWN_OUT_OF_BOUNDS_RECTANGLE.x - (size[0] * 3), SPAWN_OUT_OF_BOUNDS_RECTANGLE.x + SPAWN_OUT_OF_BOUNDS_RECTANGLE.w + (size[0] * 3)); 
			position[1] = SPAWN_OUT_OF_BOUNDS_RECTANGLE.y - (size[1] * 3);
			break;
		case SPAWN_DIRECTION_EAST:
			position[0] = SPAWN_OUT_OF_BOUNDS_RECTANGLE.x + SPAWN_OUT_OF_BOUNDS_RECTANGLE.w + (size[0] * 3);
			position[1] = RANDOM_F32(SPAWN_OUT_OF_BOUNDS_RECTANGLE.y - (size[1] * 3), SPAWN_OUT_OF_BOUNDS_RECTANGLE.y + SPAWN_OUT_OF_BOUNDS_RECTANGLE.h + (size[1] * 3)); 
			break;
		case SPAWN_DIRECTION_SOUTH:
			position[0] = RANDOM_F32(SPAWN_OUT_OF_BOUNDS_RECTANGLE.x - (size[0] * 3), SPAWN_OUT_OF_BOUNDS_RECTANGLE.x + SPAWN_OUT_OF_BOUNDS_RECTANGLE.w + (size[0] * 3)); 
			position[1] = SPAWN_OUT_OF_BOUNDS_RECTANGLE.y + SPAWN_OUT_OF_BOUNDS_RECTANGLE.h + (size[1] * 3);
			break;
		case SPAWN_DIRECTION_WEST:
			position[0] = SPAWN_OUT_OF_BOUNDS_RECTANGLE.x - (size[0] * 3);
			position[1] = RANDOM_F32(SPAWN_OUT_OF_BOUNDS_RECTANGLE.y - (size[1] * 3), SPAWN_OUT_OF_BOUNDS_RECTANGLE.y + SPAWN_OUT_OF_BOUNDS_RECTANGLE.h + (size[1] * 3)); 
			break;
	}

}

/* Spawns an entity in bounds. */
static void
_spawn_entity_in_bounds_position_get(ComponentActionSpawn* self, ECS* ecs, GameEntityType type, vec3 position, vec2 size, u32 id)
{
	position[0] = RANDOM_F32
	(
		SPAWN_IN_BOUNDS_RECTANGLE.x,
		SPAWN_IN_BOUNDS_RECTANGLE.x + SPAWN_IN_BOUNDS_RECTANGLE.w
	); 

	position[1] = RANDOM_F32
	(
		SPAWN_IN_BOUNDS_RECTANGLE.y, 
		SPAWN_IN_BOUNDS_RECTANGLE.y + SPAWN_IN_BOUNDS_RECTANGLE.h
	); 
}

/* Initialize and position the spawn enemy. */
static void
_spawn_entity(ComponentActionSpawn* self, ECS* ecs)
{
	vec3 position;
	vec2 size;
	u32 id;
	ComponentPhysics* physics;
	ComponentPhysics* targetPhysics;
	ComponentGameObject* gameObject;
	GameEntityType type;
	SpawnEntryType spawnEntryType;
	ComponentLevel* level;
	SpawnDirection direction;
	f32 angle;
	f32 speed;
	u32 usedLevel;
	IconType iconType;

	id = ecs_entity_add(ecs);

	glm_vec2_zero(size);
	glm_vec3_zero(position);

	type = _game_entity_type_get(self, ecs);

	level = ecs_component_get(ecs, ECS_COMPONENT_LEVEL, self->component.id);

	usedLevel = level->value + self->levelBonus;

	spawnEntryType = SPAWN_GAME_ENTITY_ENTRY_TYPES[type];

	switch (type)
	{
		case GAME_ENTITY_CUPPER:
			glm_vec2_copy((f32*)CUPPER_SIZE, size);
			iconType = ICON_CUPPER;
			break;
		case GAME_ENTITY_BEANIE:
			glm_vec2_copy((f32*)BEANIE_SIZE, size);
			iconType = ICON_BEANIE;
			break;
		case GAME_ENTITY_CHIPPER:
			glm_vec2_copy((f32*)CHIPPER_SIZE, size);
			iconType = ICON_CHIPPER;
			break;
		case GAME_ENTITY_CUSTARPEDO:
			glm_vec2_copy((f32*)CUSTARPEDO_SIZE, size);
			iconType = ICON_CUSTARPEDO;
			break;
		case GAME_ENTITY_CRUMBLER:
			glm_vec2_copy((f32*)CRUMBLER_SIZE, size);
			iconType = ICON_CRUMBLER;
			break;
		case GAME_ENTITY_EXCORSANT:
			glm_vec2_copy((f32*)EXCORSANT_SIZE, size);
			iconType = ICON_EXCORSANT;
			break;
		default:
			iconType = ICON_EMPTY_ONE;
			break;
	}

	switch (spawnEntryType)
	{
		case SPAWN_ENTRY_WANDER:
			_spawn_entity_out_of_bounds_position_get(self, ecs, type, position, size, id, &direction);
			_alert_direction_spawn(self, ecs, position, direction, iconType);
			break;
		case SPAWN_ENTRY_FALL:
		case SPAWN_ENTRY_EMERGE:
			_spawn_entity_in_bounds_position_get(self, ecs, type, position, size, id);
			_alert_no_direction_spawn(self, ecs, position, iconType);
			break;
		default:
			break;
	}


	switch (type)
	{
		case GAME_ENTITY_CUPPER:
			entity_cupper_init(ecs, id, position, self->targetID, usedLevel);
			break;
		case GAME_ENTITY_BEANIE:
			entity_beanie_init(ecs, id, position, self->targetID, usedLevel);
			break;
		case GAME_ENTITY_CHIPPER:
			entity_chipper_init(ecs, id, position, self->targetID, usedLevel);
			break;
		case GAME_ENTITY_CUSTARPEDO:
			entity_custarpedo_init(ecs, id, position, self->targetID, usedLevel);
			break;
		case GAME_ENTITY_CRUMBLER:
			entity_crumbler_init(ecs, id, position, usedLevel);
			break;
		case GAME_ENTITY_EXCORSANT:
			entity_excorsant_init(ecs, id, position, self->targetID, usedLevel);
			break;
		default:
			break;
	}
}

/* Ticks action_spawn component. */
void
component_action_spawn_add(ComponentActionSpawn* self, ECS* ecs)
{
	vector_init(&self->types, sizeof(GameEntityType));
	ecs_component_add(ecs, ECS_COMPONENT_LEVEL, self->component.id);
	ecs_component_add(ecs, ECS_COMPONENT_WAVE, self->component.id);
	_spawn_timer_set(self, ecs);
}

/* Ticks action_spawn component. */
void
component_action_spawn_delete(ComponentActionSpawn* self, ECS* ecs)
{
	vector_free(&self->types);
	ecs_component_delete(ecs, ECS_COMPONENT_LEVEL, self->component.id);
	ecs_component_delete(ecs, ECS_COMPONENT_WAVE, self->component.id);
}

/* Ticks action_spawn component. */
void
component_action_spawn_init(ComponentActionSpawn* self, u32 targetID)
{
	self->targetID = targetID;
}

/* Adds the currently spawning entities from a wave. */
void
component_action_spawn_wave_add(ComponentActionSpawn* self, ECS* ecs, const SpawnWave* spawnWave)
{
	ComponentWave* wave;
	bool isTypeExist;

	wave = ecs_component_get(ecs, ECS_COMPONENT_WAVE, self->component.id);

	isTypeExist = false;

	/* Only get new types from the types parameter; discard redundant ones. */
	for (s32 i = 0; i < (s32)spawnWave->typeCount; i++)
	{
		GameEntityType type;

		type = spawnWave->types[i];

		for (s32 j = 0; j < (s32)self->types.count; j++)
		{
			GameEntityType checkType;

			checkType = *(GameEntityType*)vector_get(&self->types, j);

			if (checkType == type)
			{
				isTypeExist = true;
				break;
			}
		}

		if (isTypeExist)
		{
			isTypeExist = false;
			continue;
		}

		vector_push(&self->types, &type);
	}

	self->combatTimer += spawnWave->timer;
	
	if (!self->isCombat)
		music_play(&ecs->game->resources.music[MUSIC_COMBAT_ONE], true);

	self->isCombat = true;
	self->mercyTimer = COMPONENT_ACTION_SPAWN_MERCY_TIMER_MAX;

	wave->value++;
}

/* Ticks action_spawn component. */
void
component_action_spawn_tick(ComponentActionSpawn* self, ECS* ecs)
{
	if (self->isCombat)
	{
		self->combatTimer--;
		self->timer--;
	}

	if (self->combatTimer <= 0)
	{
		self->mercyTimer--;

		if (self->mercyTimer <= 0)
		{
			for (s32 i = 0; i < (s32)ecs->lists[ECS_COMPONENT_ENEMY].components.count; i++)
			{
				ComponentEnemy* enemy;

				enemy = ecs_component_from_index_get(ecs, ECS_COMPONENT_ENEMY, i);

				ecs_entity_delete(ecs, enemy->component.id);
			}

			self->mercyTimer = COMPONENT_ACTION_SPAWN_MERCY_TIMER_MAX;
		}
	}

	if (self->combatTimer <= 0 && ecs->lists[ECS_COMPONENT_ENEMY].components.count == 0)
	{
		if (self->isCombat)
			music_play(&ecs->game->resources.music[MUSIC_CALM], true);

		vector_clear(&self->types);
		self->levelBonus = 0;
		self->isCombat = false;
		return;
	}

	if (self->timer <= 0)
	{
		if (self->combatTimer > 0)
		{
			_spawn_entity(self, ecs);
			_spawn_timer_set(self, ecs);
		}
	}

	if (self->combatTimer <= 0 )
		self->combatTimer = 0;
}
